/*
 * cc -lpciaccess drift_control945.c -o drift_control945
 *
 * drift_control:
 *   - without args simply monitors drift of frame rate against loop cycle
 *   - with args adjusts frame rate to loop cycle frequency 
 *
 * prerequisites:
 *   - Xserver must be running to initialize a proper 50Hz modeline
 *   - XV extension may NOT be actively used when using this tool
 */
#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#ifdef NO_LIBPCIACCESS
#include <fcntl.h>
#include <sys/mman.h>
#define I945G_IO_PHYS_BASE 0x50200000
#define I945G_IO_SIZE 0x80000
#else /**/
#include <pciaccess.h>
#define I945G_IO_BAR 0
#endif /**/
#define STANDALONE

/* --- 8< --- */
/*
 * field cycle duration in usecs for PAL
 */
#define SYF_PAL_FIELD_CYCLE 20000

/*
 * frame cycle duration in usecs for PAL
 */
#define SYF_PAL_FRAME_CYCLE (SYF_PAL_FIELD_CYCLE << 1)

/*
 * dependent on interlaced (progressive) input frame rate 25 (50) fps
 * we set this to 0 (1). other params should adjust accordingly.
 */
#define SYF_INPUT_DOUBLE_RATE 0

/* NOT CURRENTLY USED
 * prefilter to prevent stray updates.
 * our software PLL will not try to lock for these.
 */
#define SYF_CATCH_RANGE (SYF_PAL_FRAME_CYCLE - 500 >> SYF_INPUT_DOUBLE_RATE)

/*
 * updates outside time window defined by this
 * value spawn warnings when in debug mode
 */
#define SYF_WARN_RANGE (5000 >> SYF_INPUT_DOUBLE_RATE)

/*
 * we average 25 (50) frames to yield a cycle time of
 * about one second for analysis of frame rate data.
 * this serves as frequency divider for our software PLL.
 */
#define SYF_PLL_DIVIDER (25 << SYF_INPUT_DOUBLE_RATE)

/*
 * input frame cycle duration in usecs for 25 (50) fps
 */
#define SYF_FRAME_CYCLE (SYF_PAL_FRAME_CYCLE >> SYF_INPUT_DOUBLE_RATE)

/* NOT CURRENTLY USED
 * we slightly displace sync point from center position. this makes the
 * system less susceptible to temporary field delivery irregularities.
 */
#define SYF_SYNC_POINT_OFFSET 0

/*
 * offset in usecs from double buffer switch where we preferredly try to place double
 * buffer updates. 
 * this minimizes sensivity to jitter of our software PLL phase comparator.
 */
#define SYF_SYNC_POINT (SYF_PAL_FIELD_CYCLE - SYF_SYNC_POINT_OFFSET)

/*
 * one trim increment compensates drift speed for about 29usec/sec.
 * this represents resolution of our VCO input. 
 */
#define SYF_MIN_STEP_USEC 700

/* NOT CURRENTLY USED
 * factor weighting drift against sync point displacement
 * when calculating overall compensation 
 */
#define SYF_DISP_DRIFT_FACTOR 1

/* NOT CURRENTLY USED
 * to allow for smooth adaption also on slower devices we delimit maximum
 * trim change per step size.
 * this implements some kind of low pass filter for our VCO input.
 */
#define SYF_MAX_TRIM_REL 1000

/*
 * maximum absolute trim values allowed due to current
 * hardware/driver contraints.
 * this delimits 'maximum voltage' controlling our VCO.
 * currently allowed range is defined symmetrically around the center voltage.
 */
#define SYF_MAX_TRIM_ABS 2

#ifdef STANDALONE

#define DOVSTA          0x30008
#define HTOTAL_A        0x60000
#define HBLANK_A        0x60004
#define HSYNC_A         0x60008
#define PIPEA_DSL       0x70000
#define PIPEACONF       0x70008

#define ErrorF printf
#define B(a) (*(argv + (a)) ? *(argv + (a)) : 0)
#define INREG(reg) *(volatile unsigned *)(vptr + (reg))
#define OUTREG(reg, val) INREG(reg) = (val)
#define min(a, b) ((a) <= (b) ? (a) : (b))
#define max(a, b) ((a) >= (b) ? (a) : (b))
#define abs(a) ((a) >= 0 ? (a) : -(a))

#define VSF_SUB(a, b, c) \
    if ((a).tv_usec < (b).tv_usec) { \
        (c).tv_sec = (a).tv_sec - (b).tv_sec - 1; \
        (c).tv_usec = (a).tv_usec - (b).tv_usec + 1000000; \
    } else { \
        (c).tv_sec = (a).tv_sec - (b).tv_sec; \
        (c).tv_usec = (a).tv_usec - (b).tv_usec; \
    }

#define VSF_TV2USEC(a, b) \
    (b) = (a).tv_sec * 1000000 + (a).tv_usec;

struct _I830_s {
    int SYF_debug;
} I830 = {
    1,
}, *pI830 = &I830;

#else

#define B(a) (a)

#endif

#include <time.h>
#include <unistd.h>

#define OC_FIELD (1 << 19)

typedef struct _syf {
    int cnt; 
    int trim; 
    int warn;
    int drift; 
    int spoint;
} syf_t;

extern void log_graph(int, int);

void
log_graph(val, symb)
{
    static char headr[] = "|<- -20ms                               0                              +20ms ->|";
    static char meter[sizeof(headr)];

    if (!symb || symb == 1) {
        if (!symb) ErrorF("%s", headr);
        if (symb == 1) ErrorF("%s", meter);
        memset(meter, '-', sizeof(headr) - 1);
        return;
    }
    val = (SYF_SYNC_POINT + val) / 500;
    val = min(val, (int)sizeof(headr) - 2);
    val = max(val, 0);
    if (symb == ':') {
        meter[val] = meter[val] == '%' ? '#' : symb;
    } else if (symb == '*') {
        meter[val] = meter[val] == '%' ? '1' :
                     meter[val] == ':' ? '2' :
                     meter[val] == '#' ? '3' : symb;
    } else {
        meter[val] = symb;
    }
}
/* --- 8< --- */

main(argc, argv)
    char **argv;
{
    static int vbl_usec_prev = ~0;
    static int htotal_a, hsync_a;
    static syf_t syf;
    int dovsta, dovsta_1, pipea_dsl, pipea_dsl_1, trim, vbl_usec;

    static usleepv = SYF_FRAME_CYCLE;
    static struct timeval tv_call_prev;
    struct timeval tv_tmp, tv_call;
    void *vptr;
    int usecs;
#ifdef NO_LIBPCIACCESS
    int fd;

    if ((fd = open("/dev/mem", O_RDWR | O_SYNC)) == -1) {
        ErrorF("open: %s\n", strerror(errno)); exit(-1);
    }
    vptr = mmap(0, I945G_IO_SIZE, 
                    PROT_READ | PROT_WRITE, MAP_SHARED, fd, I945G_IO_PHYS_BASE);
    if ((long)vptr == -1) {
        ErrorF("mmap: %s\n", strerror(errno)); exit(-1);
    }
#else
    struct pci_device *dev;
    struct pci_id_match match;
    struct pci_device_iterator *iter;

    if (pci_system_init()) {
        ErrorF("pci_system_init: %s\n", strerror(errno)); exit(-1);
    }
    match.vendor_id = 0x8086;
    match.device_id = PCI_MATCH_ANY;
    match.subvendor_id = PCI_MATCH_ANY;
    match.subdevice_id = PCI_MATCH_ANY;
    match.device_class = 0;
    match.device_class_mask = 0;
    if (!(iter = pci_id_match_iterator_create(&match))) {
        ErrorF("pci_id_match_iterator_create: %s\n", strerror(errno)); exit(-1);
    }
    while (1) {
        if (!(dev = pci_device_next(iter))) {
            ErrorF("device not found\n"); exit(-1);
        }
        if (dev->device_id == 0x27ae 
         || dev->device_id == 0x2772
         || dev->device_id == 0x2592
#if 0
        if (dev->device_id == 0x2772
         || dev->device_id == 0x5d44 
#endif
                                        ) {
            break;
        }
    }
    if (pci_device_probe(dev)) {
        ErrorF("pci_device_probe: %s\n", strerror(errno)); exit(-1);
    }
    ErrorF("phys_base: 0x%08llx phys_len: 0x%08llx\n", 
                                        dev->regions[I945G_IO_BAR].base_addr, 
                                        dev->regions[I945G_IO_BAR].size);
    if (pci_device_map_range(dev, dev->regions[I945G_IO_BAR].base_addr, 
                                dev->regions[I945G_IO_BAR].size, 1, &vptr)) {
        ErrorF("pci_device_map_range: %s\n", strerror(errno)); exit(-1);
    }
#endif

    while (1) {
        gettimeofday(&tv_call, 0);
        if (tv_call_prev.tv_sec) {
            VSF_SUB(tv_call, tv_call_prev, tv_tmp);
            VSF_TV2USEC(tv_tmp, usecs);
            usleepv += SYF_FRAME_CYCLE - usecs;
        }
        tv_call_prev = tv_call;

/* --- 8< --- */
        if (!htotal_a) {
            htotal_a = INREG(HTOTAL_A);
            hsync_a = INREG(HSYNC_A);
        }
        trim = (INREG(HTOTAL_A) >> 16) - (htotal_a >> 16);

        /*
         *  the chip does not provide current field status directly.
         *  but we can derive that from xor'ed dovsta and pipea_dsl contents.
         *
         *  DOVSTA 0x80104000 PIPEA_DSL 0x00000000   0   0     0
         *  [...]
         *  DOVSTA 0x00104000 PIPEA_DSL 0x0000011e 286 286 18304
         *  DOVSTA 0x80085000 PIPEA_DSL 0x0000011f 287 287 18368
         *  [...]
         *  DOVSTA 0x80184000 PIPEA_DSL 0x00000138 312 312 19968
         *  DOVSTA 0x80184000 PIPEA_DSL 0x00000000   0 313 20032
         *  [...]
         *  DOVSTA 0x80184000 PIPEA_DSL 0x0000011e 286 599 38336
         *  DOVSTA 0x80105000 PIPEA_DSL 0x0000011f 287 600 38400
         *  [...]
         *  DOVSTA 0x80104000 PIPEA_DSL 0x00000138 312 625 40000
         *  DOVSTA 0x80104000 PIPEA_DSL 0x00000000   0   0     0
         *      _______________                 ________            
         * ____|               |_______________|         !(dovsta & OC_FIELD)
         * ____     ___________     ___________     ____
         *     |___|           |___|           |___|     pipea_dsl < DEADLN
         * ________                 _______________     
         *         |_______________|               |____ field status
         *
         *         0          287 313
         */

#define LFIELD 313
#define DEADLN 287
#define DYZONE 32
#define SHIFTV (LFIELD - DEADLN)
#define LFRAME (LFIELD << 1)

        /*
         * the next few lines implement a simple glitch detection/correction
         */
        dovsta = INREG(DOVSTA);
        pipea_dsl = INREG(PIPEA_DSL);
        dovsta_1 = INREG(DOVSTA);
        pipea_dsl_1 = INREG(PIPEA_DSL);

        if ((dovsta ^ dovsta_1) & OC_FIELD && pipea_dsl == pipea_dsl_1) {
            dovsta = ~dovsta;
        }

        /* 287 + 26 (+ 313) == 313 (626) */
        vbl_usec = ((pipea_dsl < DEADLN ^ !(dovsta & OC_FIELD)) * LFIELD + pipea_dsl + SHIFTV) % LFRAME << 6;

        if (vbl_usec_prev == ~0) {
            vbl_usec_prev = vbl_usec;
            log_graph(0, 0);
            ErrorF("      R\n");
#ifdef STANDALONE
            goto main_loop_end;
#else
            return;
#endif
        }
        if (pI830->SYF_debug) {
            if (abs(vbl_usec - vbl_usec_prev) > SYF_WARN_RANGE) {
                log_graph(vbl_usec - vbl_usec_prev, '%'); ++syf.warn;
            }
            if (abs(vbl_usec - SYF_SYNC_POINT) > SYF_WARN_RANGE) {
                log_graph(vbl_usec - SYF_SYNC_POINT, ':'); ++syf.warn;
            }
        }

#ifndef STANDALONE

        /*
         * we must delay the next double buffer update
         * until even field has been processed. this occurs
         * after scan line 286 has been passed.
         * on a sufficiently synchronized system running at a
         * sync point of about 20000 usecs this imposes
         * no additional sleep time.
         *
         *  0      even       20000      odd      40000
         *  |-------------------|-------------------|
         *    ...---sleep--->|
         *                 18368
         */
        if (SYF_PAL_FIELD_CYCLE - vbl_usec > 0) {
            usleep(SYF_PAL_FIELD_CYCLE - vbl_usec + DYZONE);
        }
#endif
        syf.spoint += vbl_usec - SYF_SYNC_POINT;
        syf.drift += vbl_usec - vbl_usec_prev;
        vbl_usec_prev = vbl_usec;
        ++syf.cnt;
        if (!(syf.cnt % SYF_PLL_DIVIDER)) {
            syf.spoint /= SYF_PLL_DIVIDER;
            syf.trim = (syf.drift + syf.spoint / SYF_DISP_DRIFT_FACTOR) / SYF_MIN_STEP_USEC;
            syf.trim = max(syf.trim, -SYF_MAX_TRIM_REL);
            syf.trim = min(syf.trim,  SYF_MAX_TRIM_REL);
            if (pI830->SYF_debug) {
                if (pI830->SYF_debug < 5 && syf.warn) {
                    time_t t;
                    struct tm *tm;

                    time(&t);
                    tm = localtime(&t);
                    ErrorF("%02d:%02d:%02d ", tm->tm_hour, tm->tm_min, tm->tm_sec);
                }
                if (pI830->SYF_debug < 5 && syf.warn || pI830->SYF_debug >= 5) {
                    log_graph(0, '+');
                    log_graph(syf.spoint, '|');
                    log_graph(syf.drift, '*');
                    log_graph(0, 1);
                    ErrorF("%7d %7d [%3d%+4d]\n", syf.drift, syf.spoint, (char)(trim & 0xff), syf.trim);
                }
            }
            syf.spoint = 0;
            syf.drift = 0;
            syf.warn = 0;
        }
        if (B(1) && syf.trim) {
            int t = (char)(trim & 0xff);

            if (syf.trim > 0) {
                t = min(t + 1,  SYF_MAX_TRIM_ABS);
                --syf.trim;
            } else {
                t = max(t - 1, -SYF_MAX_TRIM_ABS);
                ++syf.trim;
            }
            if (trim != t) {
                t <<= 16;
                OUTREG(HTOTAL_A,  htotal_a + t);
                OUTREG(HBLANK_A,  htotal_a + t);
                OUTREG(HSYNC_A,   hsync_a  + (t << 1));
                OUTREG(PIPEACONF, INREG(PIPEACONF));
            }
        }
/* --- 8< --- */

main_loop_end:
        usleep(usleepv);
    }
}

